"""
Decorator for deploy task
"""


from __future__ import with_statement
from fabric.api import task, settings
from py_mina.config import check_deploy_config
from py_mina.state import state, set_state
from py_mina.exceptions import *
from py_mina.tasks.deploy import *
from py_mina.echo import echo_task

def deploy_task(on_launch=None):
	"""
	Deploy task arguments decorator
	"""

	def deploy_task_fn(fn):
		"""
		Deploy task function decorator
		"""

		def pre_deploy():
			with settings(abort_exception=PreDeployError):
				try: 
					check_lock()
					lock()
					create_build_path()
					discover_latest_release()

					set_state('pre', True)
				except PreDeployError: 
					set_state('pre', False)


		def post_deploy():
			with settings(abort_exception=PostDeployError):
				try:
					if state.get('deploy') == True:
						move_build_to_releases()
						link_release_to_current()

					set_state('post', True)
				except PostDeployError: 
					set_state('post', False)


		def finallize_deploy():
			with settings(abort_exception=FinallizeDeployError):
				try: 
					if state.get('deploy') == True:
						cleanup_releases()
					remove_build_path()
					unlock()

					set_state('finallize', True)
				except Exception:
					set_state('finallize', False)


		def launch():
			if not on_launch == None and callable(on_launch):
				with settings(abort_exception=LaunchError):
					try: 
						on_launch()

						set_state('launch', True)
					except LaunchError:
						set_state('launch', False)


		def deploy(*args):
			"""
			Runs deploy process on remote server
				1) Pre deploy (lock, build path, latest_release)
				2) Deploy
				3) Post deploy (move to releases, link release to current)
				4) Finallize deploy (cleanup, remove build path, unlock)
				5) Launch application
				6) Show deploy stats
			"""
			
			check_deploy_config()

			with settings(colorize_errors=True):
				try:
					pre_deploy()

					with cd(fetch('build_to')), settings(abort_exception=DeployError):
						try:
							fn(*args)
							set_state('deploy', True)
						except DeployError: 
							set_state('deploy', False)

					post_deploy()
				finally:
					finallize_deploy()

			# Launch application if build succeeded
			if state.get('deploy') == True and state.get('post') == True: 
				launch()

			# Show deploy stats
			print_deploy_stats()

		# Copy __name__ and __doc__ from decorated function to decorator function
		deploy.__name__ = fn.__name__
		if fn.__doc__: deploy.__doc__ = fn.__doc__

		# Decorate with "fabric" @task decorator
		return task(deploy)

	return deploy_task_fn



"""
Utilities
"""


class _AttributeDict(dict):
    """
    Dictionary subclass enabling attribute lookup/assignment of keys/values.
    For example::
        >>> m = _AttributeDict({'foo': 'bar'})
        >>> m.foo
        'bar'
        >>> m.foo = 'not bar'
        >>> m['foo']
        'not bar'
    ``_AttributeDict`` objects also provide ``.first()`` which acts like
    ``.get()`` but accepts multiple keys as arguments, and returns the value of
    the first hit, e.g.::
        >>> m = _AttributeDict({'foo': 'bar', 'biz': 'baz'})
        >>> m.first('wrong', 'incorrect', 'foo', 'biz')
        'bar'
    """
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            # to conform with __getattr__ spec
            raise AttributeError(key)

    def __setattr__(self, key, value):
        self[key] = value

    def first(self, *names):
        for name in names:
            value = self.get(name)
            if value:
                return value



"""
Templates
"""

deploy_file_sample = '''"""
py_mina - deployfile sample
"""

from py_mina import *
from py_mina.tasks import git_clone, link_shared_paths


# Settings - remote server


set('user', 'deploy')
set('hosts', ['localhost'])


# Settings - application


set('deploy_to', '/var/www/example_application')
set('repository', 'https://github.com/py-mina-deploy/py-mina')
set('branch', 'master')
set('shared_dirs', ['logs', 'tmp'])
set('shared_files', [ 'example_config.yml' ])


# Tasks


@task
def restart():
	"""
	Restarts application on remote server
	"""

	run('echo "Restarting application"')


@deploy_task(on_launch=restart)
def deploy():
	"""
	Runs deploy process on remote server
	"""

	git_clone()
	link_shared_paths()

	run('echo "Deploying application"')

	# ... your deploy commands ...


@setup_task
def setup():
	"""
	Runs setup process on remote server
	"""

	run('echo "Deploy setup"')

	# ... your setup commands ...
'''


#!/usr/bin/env python
"""
Command line interface (CLI) for py_mina
"""


import argparse
import pkg_resources
from py_mina.cli.commands import *


try:
	version = pkg_resources.require("py_mina")[0].version
except Exception:
	version = '[error]'


def run():
	#
	# Arguments parser
	#


	parser = argparse.ArgumentParser(
		prog='py_mina',
		description='command line interface for py_mina')


	#
	# Options
	#


	parser.add_argument(
		'-v', '--version', 
		action='version', version='py_mina %s' % version)


	parser.add_argument(
		'-f', '--filename', 
		help='deployfile location',
		dest='filename', default='deploy/deploy.py')


	#
	# Commands
	#


	commands = parser.add_subparsers(
		help='available commands',
		dest='command', metavar='<command>')


	# Init


	init_parser = commands.add_parser(
		'init', 
		help='creates a sample deployfile')


	# Run


	run_parser = commands.add_parser(
		'run', 
		help='runs task from deployfile')


	run_parser.add_argument(
		'task',
		help='task to run on remote server')


	# List


	list_parser = commands.add_parser(
		'list', 
		help='lists available tasks in deployfile')


	#---------------------------------------------------------------------------


	args = vars(parser.parse_args())
	command = args.get('command')

	if command == 'init': cli_command_init(args)
	elif command == 'run': cli_command_run(args)
	elif command == 'list': cli_command_list(args)
	else: parser.print_help()


"""
Git tasks
"""


from __future__ import with_statement
import os
from fabric.api import *
from py_mina.echo import echo_subtask, echo_task
from py_mina.config import fetch, ensure


def git_clone():
	"""
	Clones repository to tmp build dir
	"""

	maybe_clone_git_repository()
	fetch_new_commits()
	use_git_branch()


def maybe_clone_git_repository():
	"""
	Clones bare git repository on first deploy
	"""

	ensure('repository')
	ensure('scm')


	scm = fetch('scm')

	with settings(warn_only=True):
		echo_subtask("Ensure git repository")

		if run('test -d %s' % scm).failed:
			echo_subtask("Cloning bare git repository")

			run('git clone {0} {1} --bare'.format(fetch('repository'), scm))


def fetch_new_commits():
	"""
	Fetches new git commits
	"""

	ensure('scm')
	ensure('repository')
	ensure('branch')

	echo_subtask("Fetching new commits")

	with cd(fetch('scm')):
		run('git fetch {0} "{1}:{1}" --force'.format(fetch('repository'), fetch('branch')))


def use_git_branch():
	"""
	Clone repository to "build_to"
	"""

	ensure('build_to')
	ensure('scm')
	ensure('branch')

	echo_subtask("Copy code from repository to build folder")

	with cd(fetch('build_to')):
		run('git clone {0} . --recursive --branch {1}'.format(fetch('scm'), fetch('branch')))
		run('rm -rf .git')




"""
Decorator for setup task
"""


from fabric.api import task, settings, hide
from py_mina.config import check_deploy_config
from py_mina.tasks.setup import *
from py_mina.echo import echo_task


def setup_task(fn):
	"""
	Setup task function decorator
	"""

	def setup(*args):
		"""
		Runs setup process on remote server
		"""

		echo_task('Running "setup" task')

		check_deploy_config()

		with settings(hide('output'), colorize_errors=True):
			create_required()
			create_shared()
			add_repo_to_known_hosts()
			add_host_to_known_hosts()

			fn(*args)

	# Copy __name__ and __doc__ from decorated function to decorator function
	setup.__name__ = fn.__name__
	if fn.__doc__: setup.__doc__ = fn.__doc__

	# Decorate with "fabric" @task decorator
	return task(setup)



"""
Setup tasks
"""


from __future__ import with_statement
import os
from py_mina.config import fetch
from fabric.api import run, settings, env, cd, hide
from py_mina.echo import echo_subtask


################################################################################
# Create required and shared 
################################################################################


def create_required():
	"""
	Creates required folders for deployment
	"""

	echo_subtask('Creating required folders')

	for required_path in ['shared', 'releases', 'tmp']:
		run('mkdir -p %s' % os.path.join(fetch('deploy_to'), required_path))


def create_shared():
	"""
	Creates shared_dirs and touches shared_files
	"""

	echo_subtask('Creating shared paths')

	with cd(fetch('shared_path')):
		for sdir in fetch('shared_dirs'):
			run('mkdir -p %s' % sdir)

		for sfile in fetch('shared_files'):
			directory, filename_ = os.path.split(sfile)

			if directory: 
				run('mkdir -p %s' % directory)

			run('touch ' + sfile)
			run('chmod g+rx,u+rwx ' + sfile)


################################################################################
# SSH known hosts
################################################################################


def add_repo_to_known_hosts():
	"""
	Adds current repo host to the known hosts
	"""

	echo_subtask('Adding repository to known hosts')

	pass
	# repo_host = fetch(:repository).split(%r{@|://}).last.split(%r{:|\/}).first
	# repo_port = /:([0-9]+)/.match(fetch(:repository)) && /:([0-9]+)/.match(fetch(:repository))[1] || '22'

	# next if repo_host == ""


def add_host_to_known_hosts():
	"""
	Adds domain(host) to the known hosts
	"""

	echo_subtask('Adding current host to known hosts')

	add_to_known_hosts(env.host_string, env.port)


def add_to_known_hosts(host, port):
	cmd = '''
	if ! ssh-keygen -H -F {0} &>/dev/null; then
		ssh-keyscan -t rsa -p {1} -H {0} >> ~/.ssh/known_hosts
	fi
	'''.format(host, port)

	with settings(hide('everything'), warn_only=True):
		run(cmd)



"""
Console output
"""


from fabric.colors import *


def echo_task(message):
	print(yellow('=====> %s' % message))


def echo_subtask(message):
	print(cyan('\n-----> %s' % message))


def echo_comment(message):
	print(blue(message))


def echo_danger(message):
	print(red(message))




"""
Deploy tasks
"""

from __future__ import with_statement
import os
from fabric.colors import red, green
from fabric.api import *
from py_mina.config import fetch, ensure, set
from py_mina.state import state
from py_mina.echo import *


################################################################################
# Lockfile
################################################################################


def check_lock():
	"""
	Aborts deploy if lock file is found.
	"""

	ensure('deploy_to')
	
	with settings(hide('warnings'), warn_only=True), cd(fetch('deploy_to')):
		if run('test -f deploy.lock').succeeded:
			abort("Another deploy in progress")


def lock():
	"""
	Locks deploy. Parallel deploys are not allowed.
	"""

	ensure('build_to')

	echo_subtask("Creating deploy.lock")

	with cd(fetch('deploy_to')):
		run('touch deploy.lock')


def unlock():
	"""
	Forces a deploy unlock.
	"""

	ensure('deploy_to')

	echo_subtask("Removing deploy.lock")

	with cd(fetch('deploy_to')):
		run('rm -f deploy.lock')


################################################################################
# Deploy
################################################################################


def discover_latest_release():
	"""
	Connects to remote server and discovers next release number
	"""

	releases_path = fetch('releases_path')

	echo_subtask("Discovering latest release number")

	# Get release number
	with settings(hide('warnings'), warn_only=True):
		if run('test -d %s' % releases_path).failed:
			run('mkdir -p %s' % releases_path)
			last_release_number = 0
		else:
			releases_respond = run('ls -1p %s | sed "s/\///g"' % releases_path)
			releases_dirs = list(filter(lambda x: len(x) != 0, releases_respond.split('\r\n')))

			if len(releases_dirs) > 0:
				last_release_number = max(map(lambda x: int(x), releases_dirs)) or 0
			else:
				last_release_number = 0

	release_number = last_release_number + 1

	# Set release info to config 
	set('release_number', release_number)
	set('release_to', os.path.join(releases_path, str(release_number)))


def create_build_path():
	"""
	Creates tmp dir for building app. 

	Folder will be:
		* deleted -> if build fails
		* moved to releases -> if build succeeds
	"""
	
	ensure('build_to')

	echo_subtask("Creating build path")

	with settings(warn_only=True):
		build_path = fetch('build_to')

		if run('test -d %s' % build_path).failed:
			run('mkdir -p %s' % build_path)


def link_shared_paths():
	"""
	Links paths set in shared_dirs and shared_files.
	"""

	ensure('build_to')
	ensure('shared_path')
	ensure('shared_dirs')
	ensure('shared_files')

	echo_subtask("Linking shared paths")

	with cd(fetch('build_to')):
		shared = fetch('shared_path')

		# Shared dirs
		for sdir in fetch('shared_dirs'):
			relative_path = os.path.join('./', sdir)
			directory, filename_ = os.path.split(relative_path)
			shared_path = os.path.join(shared, sdir)
		
			run('mkdir -p %s' % directory)
			run('rm -rf %s' % relative_path)
			run('ln -s "{0}" "{1}"'.format(shared_path, relative_path))

		# Shared files
		for sfile in fetch('shared_files'):
			shared_path = os.path.join(shared, sfile)
			relative_path = os.path.join('./', sfile)
			directory, filename_ = os.path.split(relative_path)

			run('mkdir -p %s' % directory)
			run('ln -sf "{0}" "{1}"'.format(shared_path, relative_path))


################################################################################
# POST Deploy
################################################################################


def move_build_to_releases():
	"""
	Moves tmp build folder to releases dir
	"""

	ensure('build_to')
	ensure('release_to')

	echo_subtask("Moving build to releases")

	run('mv %s %s' % (fetch('build_to'), fetch('release_to')))



def link_release_to_current():
	"""
	Links shared paths to build dir
	"""

	ensure('release_to')
	ensure('current_path')

	echo_subtask("Linking release to current")

	release_to = fetch('release_to')

	with cd(release_to):
		run('ln -nfs %s %s' % (release_to, fetch('current_path')))

def remove_build_path():
	"""
	Removes a temporary build dir
	"""
	ensure('build_to')
	
	echo_subtask("Removing build path")

	with settings(hide('stdout', 'warnings'), warn_only=True):
		run('rm -rf %s' % fetch('build_to'))


def cleanup_releases():
	"""
	Clean up old releases.
	"""

	ensure('releases_path')
	ensure('keep_releases')

	releases_count = str(fetch('keep_releases'))

	echo_subtask("Cleaning up old realeses. Keeping latest %s" % releases_count)

	with cd(fetch('releases_path')):
		cmd = '''
count=$(ls -A1 | sort -rn | wc -l)
remove=$((count > %s ? count - %s : 0))
ls -A1 | sort -rn | tail -n $remove | xargs rm -rf {}
		''' % (releases_count, releases_count)

		run(cmd)


def print_deploy_stats():
	release_number = fetch('release_number')
	deploy_successful = state.get('deploy') == True and state.get('post') == True

	# Status
	print('\n')
	if deploy_successful: 
		print(green('Deploy finished.'))
	else: 
		print(red('Deploy failed.'))

	# Subtasks
	if fetch('verbose') or not deploy_successful:
		print('\n')
		echo_comment('[SUBTASKS]')
		echo_comment('* pre_deploy: %s' % state.get('pre'))
		echo_comment('* deploy: %s' % state.get('deploy'))
		echo_comment('* post_deploy: %s' % state.get('post'))
		echo_comment('* finallize: %s' % state.get('finallize'))
		echo_comment('* launch: %s' % state.get('launch'))

	# Release
	if deploy_successful:
		print('\n')
		echo_comment('[RELEASE]')
		echo_comment('* number: %s' % release_number)
		echo_comment('* path: %s' % fetch('release_to'))


################################################################################
# Rollback release
################################################################################


def rollback_release():
	"""
	Rollbacks the latest release.
	"""
	
	ensure('releases_path')
	ensure('current_path')

	echo_task("Rollbacking current release")

	releases_path = fetch('releases_path')

	with cd(releases_path):
		echo_subtask('Finding previous release for rollback')
		rollback_release = run('ls -1A | sort -n | tail -n 2 | head -n 1')

		if int(rollback_release) > 0:
			echo_subtask('Linking previous release to current')
			run('ln -nfs %s/%s %s' % (releases_path, rollback_release, fetch('current_path')))

			echo_subtask('Finding current release')
			current_release = run('ls -1A | sort -n | tail -n 1')

			if int(current_release) > 0:
				echo_subtask('Removing current release')
				run('rm -rf %s/%s' % (releases_path, current_release))
			else:
				abort('Can\'t find current release for remove.')
				
		else:
			abort('Can\'t find previous release for rollback.')



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

try:
    long_description = open("README.md").read()
except IOError:
    long_description = "Deployer and server automation tool."

setup(
    name="py_mina",
    version="0.1.4",
    description="Python library for deploying applications on remote server",
    long_description=long_description,
    license="MIT",
    author="Cidevant Von Goethe",
    author_email="cidevant@mail.ru",
    url="https://github.com/py-mina-deploy/py-mina",
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'py_mina = py_mina.cli.cli:run',
        ]
    },
    install_requires=[
		'fabric3',
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
    ]
)



"""
Configuration manager
"""


import os
import time
from fabric.api import env
from py_mina.utils import _AttributeDict
import py_mina.state
from py_mina.exceptions import FetchConfigError, EnsureConfigError, BadConfigError


################################################################################
# Default config
################################################################################


config = _AttributeDict({
	'keep_releases': 5,
	'shared_files': [],
	'shared_dirs': [],
	'abort_on_prompts': False,
	'verbose': False,
	'farbric_config_settings': ['user', 'hosts', 'abort_on_prompts']
})


################################################################################
# Set config
################################################################################


def set(key, value):
	"""
	Set config setting.
	"""

	if key in config.get('farbric_config_settings'):
		env.update({ key: value })
		config.update({ key: value })
	elif key == 'deploy_to':
		config.update({
			'deploy_to': value, 
			'scm': os.path.join(value, 'scm'), 
			'shared_path': os.path.join(value, 'shared'), 
			'current_path': os.path.join(value, 'current'), 
			'releases_path': os.path.join(value, 'releases'), 
			'build_to': os.path.join(value, 'tmp', 'build-' + str(time.time())),
		})
	else:
		config.update({ key: value })


# Alias to prevent conflict when importing "py_mina.config" and "py_mina.state"
set_config = set


################################################################################
# Helpers
################################################################################


def check_deploy_config():
	"""
	Ensures required settings is config for deploy and setup tasks
	"""

	required_settings = ['user', 'hosts', 'deploy_to', 'repository', 'branch']
	try:
		for setting in required_settings:
			ensure(setting)
	except EnsureConfigError:
		msg = '''
Bad config! 
Required settings: {0}
Current config: {1}
		'''.format(required_settings, config)

		raise BadConfigError(msg)


def fetch(key, default_value=None):
	"""
	Fetches config setting by given key.
	Returns default_value if key is missing and default value is provided.
	"""

	if key in config.keys():
		return config.get(key)
	else:
		if default_value != None:
			return default_value
		else:
			raise FetchConfigError('"%s" is not defined' % key)


def ensure(key):
	"""
	Ensures that config contains value for given key.
	"""

	if not key in config.keys():
		raise EnsureConfigError('"%s" must be defined' % key)



"""
py_mina CLI commands
"""


import os
from fabric.main import load_fabfile
from fabric.tasks import execute
from fabric.colors import green, red
from py_mina.cli.templates import deploy_file_sample


#
# List
#


def cli_command_list(args):
	"""
	Runs task from deploy file
	"""
	
	filepath = args.get('filename')
	task_name = args.get('task')

	if filepath:
		fullpath = os.path.join(os.path.realpath(os.curdir), filepath)
		docstring, callables, default = load_fabfile(fullpath)

		if not docstring == None:
			print(docstring.rstrip('\n'))

		print('\nAvailable commands:\n')

		for task_name in callables:
			task_desc = callables.get(task_name).__doc__

			if not task_desc == None:
				description = task_desc.lstrip('\n').rstrip('\n')
			else:
				description = '\n'

			print('%s   %s' % (green(task_name), description))


#
# Run
#


def cli_command_run(args):
	"""
	Runs task from deploy file
	"""
	
	filepath = args.get('filename')
	task_name = args.get('task')

	if filepath:
		fullpath = os.path.join(os.path.realpath(os.curdir), filepath)
		docstring, callables, default = load_fabfile(fullpath)

		if task_name in callables:
			execute(callables[task_name])


#
# Init
#


def cli_command_init(args):
	"""
	Creates a sample config file (default path is ./deploy/deploy.py)
	"""

	filepath = args.get('filename')

	if filepath:
		fullpath = os.path.join(os.path.realpath(os.curdir), filepath)
		directory, deployfile = os.path.split(fullpath)

		try:
			# Create deployfile directory if not exists
			if not os.path.exists(directory):
			    os.makedirs(directory)

			# Write template string to file
			with open(filepath, 'w') as f:
				f.write(deploy_file_sample)

			print(green('Deployfile successfully created in %s' % fullpath))

		except Exception as error: 
			print(red(error))